@@ -39,7 +39,14 @@ from abc import abstractmethod, ABC
 import joblib
 import numpy as np
 import tensorflow as tf
-from vegasflow.configflow import MAX_EVENTS_LIMIT, DEFAULT_ACTIVE_DEVICES, DTYPE
+from vegasflow.configflow import (
+    MAX_EVENTS_LIMIT,
+    DEFAULT_ACTIVE_DEVICES,
+    DTYPE,
+    TECH_CUT,
+    float_me,
+    fone,
+)
 
 import logging
 
@@ -47,8 +54,8 @@ logger = logging.getLogger(__name__)
 
 
 def print_iteration(it, res, error, extra="", threshold=0.1):
-    """ Checks the size of the result to select between
-    scientific notation and floating point notation """
+    """Checks the size of the result to select between
+    scientific notation and floating point notation"""
     # note: actually, the flag 'g' does this automatically
     # but I prefer to choose the precision myself...
     if res < threshold:
@@ -58,7 +65,7 @@ def print_iteration(it, res, error, extra="", threshold=0.1):
 
 
 def _accumulate(accumulators):
-    """ Accumulate all the quantities in accumulators
+    """Accumulate all the quantities in accumulators
     The default accumulation is implemented for tensorflow tensors
     as a sum of all partial results.
 
@@ -94,6 +101,7 @@ class MonteCarloFlow(ABC):
             Note: for a better performance, when n_events is greater than the event limit,
             `n_events` should be exactly divisible by `events_limit`
         `list_devices`: list of device type to use (use `None` to do the tensorflow default)
+        `simplify_signature`: if true only the array of random numbers will be passed to the integrand
     """
 
     def __init__(
@@ -103,18 +111,21 @@ class MonteCarloFlow(ABC):
         events_limit=MAX_EVENTS_LIMIT,
         list_devices=DEFAULT_ACTIVE_DEVICES,
         verbose=True,
+        simplify_signature=False,
     ):
         # Save some parameters
         self.n_dim = n_dim
-        self.xjac = 1.0 / n_events
         self.integrand = None
         self.event = None
+        self.simplify_signature = simplify_signature
         self._verbose = verbose
         self._history = []
         self.n_events = n_events
+        self._xjac = float_me(1.0 / n_events)
         self._events_per_run = min(events_limit, n_events)
-        self.lock = threading.Lock()
+        self.distribute = False
         if list_devices:
+            self.lock = threading.Lock()
             # List all devices from the list that can be found by tensorflow
             devices = []
             for device_type in list_devices:
@@ -131,13 +142,13 @@ class MonteCarloFlow(ABC):
 
     @property
     def events_per_run(self):
-        """ Number of events to run in a single step.
+        """Number of events to run in a single step.
         Use this variable to control how much the memory will be loaded"""
         return self._events_per_run
 
     @events_per_run.setter
     def events_per_run(self, val):
-        """ Set the number of events per single step """
+        """Set the number of events per single step"""
         self._events_per_run = min(val, self.n_events)
         if self.n_events % self._events_per_run != 0:
             logger.warning(
@@ -147,7 +158,7 @@ class MonteCarloFlow(ABC):
 
     @property
     def history(self):
-        """ Returns a list with a tuple of results per iteration
+        """Returns a list with a tuple of results per iteration
         This tuple contains:
 
         - `result`: result of each iteration
@@ -158,23 +169,71 @@ class MonteCarloFlow(ABC):
         """
         return self._history
 
+    def generate_random_array(self, n_events):
+        """External interface for the generation of random
+        points as a 2D array of (n_events, n_dim).
+        It calls the internal version of ``_generate_random_array``
+
+        Parameters
+        ----------
+            `n_events`: number of events to generate
+
+
+        Returns
+        -------
+            `rnds`: array of (n_events, n_dim) random points
+            `idx` : index associated to each random point
+            `p(x)` : p(x) associated to the random points
+        """
+        rnds, idx, xjac_raw = self._generate_random_array(n_events)
+        # returns a p(x) corresponding to the number of events
+        # the algorithm was trained with, reweight
+        xjac = xjac_raw / self._xjac / n_events
+        return rnds, idx, xjac
+
+    def _generate_random_array(self, n_events):
+        """Generate a 2D array of (n_events, n_dim) points
+        For the weight of the given point, this function is considered
+        as part of an integration with ``self.n_events`` calls.
+
+        Parameters
+        ----------
+            `n_events`: number of events to generate
+
+        Returns
+        -------
+            `rnds`: array of (n_events, n_dim) random points
+            `idx` : index associated to each random point
+            `wgt` : wgt associated to the random point
+        """
+        rnds = tf.random.uniform(
+            (n_events, self.n_dim), minval=TECH_CUT, maxval=1.0 - TECH_CUT, dtype=DTYPE
+        )
+        idx = 0
+        return rnds, idx, self._xjac
+
     #### Abstract methods
     @abstractmethod
     def _run_iteration(self):
-        """ Run one iteration (i.e., `self.n_events`) of the
-        Monte Carlo integration """
+        """Run one iteration (i.e., `self.n_events`) of the
+        Monte Carlo integration"""
 
     @abstractmethod
     def _run_event(self, integrand, ncalls=None):
-        """ Run one single event of the Monte Carlo integration
-        the output must be a tuple """
+        """Run one single event of the Monte Carlo integration
+        the output must be a tuple"""
         result = self.event()
         return result, pow(result, 2)
 
+    #### Integration management
+    def set_seed(self, seed):
+        """Sets the interation seed"""
+        tf.random.set_seed(seed)
+
     #### Device management methods
     def get_device(self):
-        """ Looks in the list of devices until it finds a device available, once found
-        makes the device unavailable and returns it """
+        """Looks in the list of devices until it finds a device available, once found
+        makes the device unavailable and returns it"""
         use_dev = None
         self.lock.acquire()
         try:
@@ -189,7 +248,7 @@ class MonteCarloFlow(ABC):
         return use_dev
 
     def release_device(self, device):
-        """ Makes `device` available again """
+        """Makes `device` available again"""
         self.lock.acquire()
         try:
             self.devices[device] = True
@@ -197,7 +256,7 @@ class MonteCarloFlow(ABC):
             self.lock.release()
 
     def device_run(self, ncalls, sent_pc=100.0, **kwargs):
-        """ Wrapper function to select a specific device when running the event
+        """Wrapper function to select a specific device when running the event
         If the devices were not set, tensorflow default will be used
 
         Parameters
@@ -221,6 +280,27 @@ class MonteCarloFlow(ABC):
             result = self.event(ncalls=ncalls, **kwargs)
         return result
 
+    def set_distribute(self, queue_object):
+        """Uses dask to distribute the vegasflow run onto a cluster
+        Takes as input a queue_object defining the jobs to be sent
+
+        Parameters
+        ----------
+            `queue_object`: dask_jobqueue object
+        """
+        try:
+            import dask.distributed  # pylint: disable=import-error
+        except ImportError:
+            raise ImportError("Install dask and distributed to use `set_distribute`")
+        if self.devices is not None:
+            logger.warning("`set_distribute` overrides any previous device configuration")
+        self.list_devices = None
+        self.lock = None
+        self.pool = None
+        self.devices = None
+        self.cluster = queue_object
+        self.distribute = True
+
     def run_event(self, **kwargs):
         """
         Runs the Monte Carlo event. This corresponds to a number of calls
@@ -258,6 +338,20 @@ class MonteCarloFlow(ABC):
                 delay_job = joblib.delayed(self.device_run)(ncalls, sent_pc=pc, **kwargs)
                 running_pool.append(delay_job)
             accumulators = self.pool(running_pool)
+        elif self.distribute:
+            from dask.distributed import Client  # pylint: disable=import-error
+
+            cluster = self.cluster
+            self.cluster = None  # the cluster might not be pickable # TODO
+            # Generate the client to control the distribution using the cluster variable
+            client = Client(cluster)
+            accumulators_future = client.map(self.device_run, events_to_do, percentages)
+            result_future = client.submit(_accumulate, accumulators_future)
+            result = result_future.result()
+            # Liberate the client
+            client.close()
+            self.cluster = cluster
+            return result
         else:
             accumulators = []
             for ncalls, pc in zip(events_to_do, percentages):
@@ -266,7 +360,7 @@ class MonteCarloFlow(ABC):
         return _accumulate(accumulators)
 
     def compile(self, integrand, compilable=True):
-        """ Receives an integrand, prepares it for integration
+        """Receives an integrand, prepares it for integration
         and tries to compile unless told otherwise.
 
         The input integrand must receive, as an input, an array of random numbers.
@@ -284,6 +378,11 @@ class MonteCarloFlow(ABC):
 
         - `integrand(array_random, **kwargs)`.
 
+        if the integrator is instantiated with the ``simplify_signature`` argument
+        the signature will be:
+
+        - `integrand(array_random)`
+
         Parameters
         ----------
             `integrand`: the function to integrate
@@ -291,8 +390,19 @@ class MonteCarloFlow(ABC):
                 is not passed through `tf.function`
         """
         self.integrand = integrand
+        compile_options = {"experimental_autograph_options": tf.autograph.experimental.Feature.ALL}
+
         if compilable:
-            tf_integrand = tf.function(integrand)
+            if self.simplify_signature:
+                compile_options["input_signature"] = [
+                    tf.TensorSpec(shape=[None, self.n_dim], dtype=DTYPE)
+                ]
+            # Don't override user own compilation
+            try:
+                integrand.function_spec
+                tf_integrand = integrand
+            except AttributeError:
+                tf_integrand = tf.function(integrand, **compile_options)
         else:
             tf_integrand = integrand
 
@@ -305,7 +415,7 @@ class MonteCarloFlow(ABC):
             self.event = run_event
 
     def run_integration(self, n_iter, log_time=True, histograms=None):
-        """ Runs the integrator for the chosen number of iterations.
+        """Runs the integrator for the chosen number of iterations.
 
         `histograms` must be a tuple of tf.Variables.
         At the end of all iterations the histograms per iteration will
@@ -385,11 +495,11 @@ class MonteCarloFlow(ABC):
         final_result = aux_res / weight_sum
         sigma = np.sqrt(1.0 / weight_sum)
         logger.info(f" > Final results: {final_result.numpy():g} +/- {sigma:g}")
-        return final_result, sigma
+        return final_result.numpy(), sigma
 
 
 def wrapper(integrator_class, integrand, n_dim, n_iter, total_n_events, compilable=True):
-    """ Convenience wrapper
+    """Convenience wrapper for integration
 
     Parameters
     ----------
@@ -407,3 +517,35 @@ def wrapper(integrator_class, integrand, n_dim, n_iter, total_n_events, compilab
     mc_instance = integrator_class(n_dim, total_n_events)
     mc_instance.compile(integrand, compilable=compilable)
     return mc_instance.run_integration(n_iter)
+
+
+def sampler(
+    integrator_class,
+    integrand,
+    n_dim,
+    total_n_events,
+    training_steps=5,
+    compilable=True,
+    return_class=False,
+):
+    """Convenience wrapper for sampling random numbers
+
+    Parameters
+    ----------
+        `integrator_class`: MonteCarloFlow inherited class
+        `integrand`: tf.function
+        `n_dim`: number of dimensions
+        `n_events`: number of events per iteration
+        `training_steps`: number of training_iterations
+        `return_class`: whether to return the full instance of the class or only the random method
+
+    Returns
+    -------
+        `sampler`: a reference to the generate_random_array method of the integrator class
+    """
+    mc_instance = integrator_class(n_dim, total_n_events, verbose=False)
+    mc_instance.compile(integrand, compilable=compilable)
+    _ = mc_instance.run_integration(training_steps, log_time=False)
+    if return_class:
+        return mc_instance
+    return mc_instance.generate_random_array
