@@ -3,13 +3,13 @@
 """
 
 from vegasflow.configflow import DTYPE, fone, fzero
-from vegasflow.monte_carlo import MonteCarloFlow, wrapper
+from vegasflow.monte_carlo import MonteCarloFlow, wrapper, sampler
 import tensorflow as tf
 
 
 class PlainFlow(MonteCarloFlow):
     """
-        Simple Monte Carlo integrator.
+    Simple Monte Carlo integrator.
     """
 
     def _run_event(self, integrand, ncalls=None):
@@ -18,10 +18,8 @@ class PlainFlow(MonteCarloFlow):
         else:
             n_events = ncalls
 
-        # Jacobian
-        xjac = 1.0 / self.n_events
         # Generate all random number for this iteration
-        rnds = tf.random.uniform((n_events, self.n_dim), minval=0, maxval=1, dtype=DTYPE)
+        rnds, _, xjac = self._generate_random_array(n_events)
         # Compute the integrand
         tmp = integrand(rnds, n_dim=self.n_dim, weight=xjac) * xjac
         tmp2 = tf.square(tmp)
@@ -39,6 +37,10 @@ class PlainFlow(MonteCarloFlow):
         return res, sigma
 
 
-def plain_wrapper(*args):
+def plain_wrapper(*args, **kwargs):
     """ Wrapper around PlainFlow """
-    return wrapper(PlainFlow, *args)
+    return wrapper(PlainFlow, *args, **kwargs)
+
+def plain_sampler(*args, **kwargs):
+    """ Wrapper sampler around PlainFlow """
+    return sampler(PlainFlow, *args, **kwargs)
